# -*- coding: utf-8 -*-
"""Python package to support the research during
the annual Lehrexkursion at Universität Hamburg.
"""
from . import csv
from . import math
from . import plots


__all__ = [
    'csv',
    'math',
    'plots',
]

__version__ = '0.5'



# -*- coding: utf-8 -*-
"""Mathematical functions.
"""
import collections
import numpy as np


__all__ = [
    'moving_average',
    'block_average',
    'bootstrap',
    'correlation',
    'rmse',
    'compare_arrays',
]


def moving_average(x, y, N, mode='same'):
    """Calculate running mean for given timeseries.

    Parameters:
        x (np.ndarray): x data.
        y (np.ndarray): y data.
        N (int): Window size.
        mode (str): Convolve mode 'valid' or 'same'.

    Returns:
        np.ndarray, np.ndarray: Adjusted x data, Averaged y data.

    """
    if mode == 'valid':
        l, t = N//2, -N//2 + 1
        x = x[l:t]
    elif mode == 'full':
        raise Exception(
            'Mode "full" is not supported to prevent boundary effects.'
            'If you know what you are doing, use numpy.convolve to use'
            'full mode.')

    return x, np.convolve(y, np.ones((N,)) / N, mode=mode)


def block_average(x, y, N):
    """Calculate the average over windows of size N.

    Parameters:
        x (np.ndarray): x data.
        y (np.ndarray): y dat.
        N (int): Window size to average.

    Returns:
        np.ndarray, np.ndarray: Every n-th x value, Corresponding averages.
    """
    x = x[N-1::N]
    y = np.array([np.nanmean(v) for v in np.split(y, y.size / N)])

    return x, y


def bootstrap(x, size=None):
    """Generate a bootstrap of given size from x.

    Note:
        This function is a simple wrapper for np.random.choice

    Parameters:
        a (1-D array-like or int):
            If an ndarray, a random sample is generated from its elements.
            If an int, the random sample is generated as if a was np.arange(n)
        size (int or tuple of ints): Output shape.
            If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  Default is None, in which case a
            single value is returned.

    Returns:
        np.ndarray, shape (size,): The generated random samples.
    """
    return np.random.choice(x, size=size, replace=True)


def rmse(x, y):
    """Calculate the Root Mean Squared Error.

    Parameters:
        x (np.ndarray): First data array.
        y (np.ndarray): Second data array.

    Returns:
        float: RMSE
    """
    return np.sqrt(np.nanmean((x - y)**2))


def correlation(x, y):
    """Return single correlatin coefficient.

    Parameters:
        x (np.ndarray): First data array.
        y (np.ndarray): Second data array.

    Returns:
        float: Correlation coefficient.
    """
    x = np.ma.masked_invalid(x)
    y = np.ma.masked_invalid(y)
    return np.ma.corrcoef(x, y)[0, 1]


def compare_arrays(x, y, verbose=False):
    """Perform a simple statistic comparison of two arrays.

    Parameters:
        x (np.ndarray): First data array.
        y (np.ndarray): Second data array.
        verbose (bool): Print output.

    Returns:
        nametuple: Statisticl comparison of input arrays.
    """

    ArrayComparison = collections.namedtuple(
        'ArrayComparison',
        ['rmse',
         'corrcoef',
         'mean_first',
         'mean_second',
         ])

    stats = ArrayComparison(
        rmse=rmse(x, y),
        corrcoef=correlation(x, y),
        mean_first=np.nanmean(x),
        mean_second=np.nanmean(y),
        )

    if verbose:
        print(
            'RMSE: {:5.3f}\n'
            'Correlation: {:5.3f}\n'
            'First Array Mean: {:5.3f}\n'
            'Second Array Mean: {:5.3f}\n'
            .format(*stats)
            )

    return stats



# -*- coding: utf-8 -*-
"""Load CSV files stored in RAW format.
"""
import numpy as np
from matplotlib import dates


__all__ = [
    'read_raw',
]


def _get_names(filename, delimiter=';', comments='%'):
    """Get variable names from CSV file in RAW format.

    Parameters:
        filename (str): Path to CSV file.
        delimiter (str): The string used to separate values.
        comments (str): The character used to indicate the start of a comment.

    Returns:
        list[str]: Variable names.
    """
    for line in open(filename, 'rt', encoding='UTF-8', errors='ignore'):
        if not line.startswith(comments):
            return line.strip().split(delimiter)


def _get_mpl_start(filename):
    with open(filename) as f:
        return dates.datestr2num(':'.join(f.readline().split(':')[1:]))


def _decimal_converter(filename, comments='%'):
    """Convert decimal separator from comma to period.

    While doing so also ignore weirdly encoded characters.

    Parameters:
        filename (str): Path to CSV file.
        comments (str): The character used to indicate the start of a comment.

    Returns:
        iterator: Return byte objects that can be read by `np.genfromtxt`.
    """
    for line in open(filename, 'rt', encoding='UTF-8', errors='ignore'):
        if line.startswith(comments):
            yield line.encode()
        else:
            yield line.replace(',', '.').encode()


def read_raw(filename, datarate=20, delimiter=';', comments='%',
             variables=None, output=None, **kwargs):
    """Read file in RAW format.

    Parameters:
        filename (str): Path to CSV file.
        datarate (float): Measurement rate [Hz].
        delimiter (str): The string used to separate values.
        comments (str): The character used to indicate the start of a comment.
        filling_values (float): Value to use for missing values.
            If `None`, use the `DefaultValue` specified in file header.
        variables (List[str]): List of variables to extract.
        output (dict): Dictionary that is updated with read data.
        **kwargs: Additional keyword arguments passed to `np.genfromtxt`.

    Returns:
        dict: Dictionary containing the data arrays.
    """
    names = _get_names(filename, delimiter, comments)
    if variables is None:
        variables = names

    data = np.genfromtxt(
        _decimal_converter(filename, comments),
        delimiter=';',
        comments=comments,
        names=names,
        usecols=variables,
        skip_header=2,
        **kwargs,
        )

    # Convert structured array to dictionary.
    if output is None:
        output = {var: data[var] for var in variables}
    else:
        for var in variables:
            output[var] = data[var]

    # TODO: Parsing the actual data rate from file.
    # Create time array using start time and data rate.
    start = _get_mpl_start(filename)  # Start time in MPL format.
    N = data.shape[0]  # Number of measurements.
    dt = 1 / (24 * 60 * 60 * datarate)  # Time difference betwen measurements.
    output['MPLTIME'] = np.arange(start, start + N * dt, dt)

    return output



# -*- coding: utf-8 -*-
"""Plot basic atmospheric properties conveniently.
"""
from matplotlib import dates
import matplotlib.pyplot as plt


__all__ = [
    'set_date_axis',
    'timeseries',
    'timeseries2d',
    'heatmap',
]


def set_date_axis(ax=None, dateformat='%d.%m.'):
    """Set DateFormatter for given AxesSubplot.

    Parameters:
        ax (AxesSubplot): Matplotlib axes.
        dateformat (str): Format string for date labels on xaxis.
    """
    if ax is None:
        ax = plt.gca()

    formatter = dates.DateFormatter(dateformat)
    ax.xaxis.set_major_formatter(formatter)
    ax.grid('on', which='both', linestyle='-')


def timeseries(time, data, xlabel='Datum', ylabel='', dateformat='%d.%m.',
               ax=None, **kwargs):
    """Create a basic timeseries plot.

    Parameters:
        time (nd.array): Date array in matplotlib time format.
        data (nd.array): Data array.
        xlabel (str): x label.
        ylabel (str): y label.
        dateformat (str): Format string for date labels on xaxis.
        ax (AxesSubplot): Matplotlib axes.
        **kwargs: Additional keyword arguments passed to `plt.plot`.

    Returns:
        Line2D: A line.

    """
    if ax is None:
        ax = plt.gca()

    line, = ax.plot(time, data, **kwargs)
    set_date_axis(ax, dateformat)

    ax.set_xlim(time.min(), time.max())
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.legend()

    return line


def timeseries2d(time, height, data, xlabel='Datum', ylabel='Höhe [m]',
                 zlabel='', dateformat='%d.%m.', rasterized=True, ax=None,
                 **kwargs):
    """Create a basic 2D timeseries plot.

    Parameters:
        time (nd.array): Date array in matplotlib time format.
        height (nd.array): Height array.
        data (nd.array): Data array.
        xlabel (str): x label.
        ylabel (str): y label.
        zlabel (str): Colorbar label.
        dateformat (str): Format string for date labels on xaxis.
        ax (AxesSubplot): Matplotlib axes.
        **kwargs: Additional keyword arguments passed to `plt.plot`.

    Returns:
        Line2D: A line.

    """
    if ax is None:
        ax = plt.gca()

    def_kwargs = {
        'rasterized': rasterized,
        'cmap': 'viridis',
        }

    def_kwargs.update(kwargs)

    ret = ax.pcolormesh(time, height, data, **kwargs)

    cb = ax.get_figure().colorbar(ret)
    cb.set_label(zlabel)

    set_date_axis(ax, dateformat)
    ax.set_xlim(time.min(), time.max())
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return ret


def heatmap(x, y, bins=20, zlabel='Anzahl', rasterized=True,
            ax=None, **kwargs):
    """Plot a heatmap of given data.

    Parameters:
        x (np.ndarray): x data.
        y (np.ndarray): y data.
        bins (None | int | [int, int] | array_like | [array, array]):

            The bin specification:

                - If int, the number of bins for the two dimensions
                (nx=ny=bins).

                - If [int, int], the number of bins in each dimension
                (nx, ny = bins).

                - If array_like, the bin edges for the two dimensions
                (x_edges=y_edges=bins).

                - If [array, array], the bin edges in each dimension
                (x_edges, y_edges = bins).

            The default value is 20.
        zlabel (str): Colobar label.
        ax (AxesSubplot): Axes to plot in.
        **kwargs: Additional keyword argumens passed to `plt.hist2d`.

    Returns:
        Return values of `plt.hist2d`: counts, xedges, yedges, Image.
    """
    if ax is None:
        ax = plt.gca()

    def_kwargs = {
        'bins': bins,
        'rasterized': rasterized,
        'cmap': plt.get_cmap('magma_r', 10),
        }

    def_kwargs.update(kwargs)

    ret = ax.hist2d(x, y, **def_kwargs)

    cb = ax.get_figure().colorbar(ret[3])
    cb.set_label(zlabel)

    return ret



# -*- coding: utf-8 -*-
"""Load CSV files stored in Wettermast format.

Attributes:
    VARIABLE_DTYPES (dict): Stores the dtype related to a variable name.
"""
import re

from matplotlib.dates import strpdate2num
import numpy as np


__all__ = [
    'read',
    'read_profile',
    'read_scat',
    'write',
]

VARIABLE_DTYPES = {
    'DATE': '<U10',
    'TIME': '<U5',
}


def _get_mpl_date(dates, fmt='%d.%m.%Y %H:%M'):
    """Convert date strings into matplotlib time format.

    Parameters:
        dates (np.array): Array containing date strings.
        fmt (str): Date string format [0].

    [0] http://pubs.opengroup.org/onlinepubs/009695399/functions/strptime.html

    Returns:
        np.array: Matplotlib time values.
    """
    return np.array([strpdate2num(fmt)(d) for d in dates])


def _get_default_value(filename):
    """Get default value for missing data from CSV file header.

    Parameters:
        filename (str): Path to CSV file.

    Returns:
        float or str: Default value for mising values.
    """
    with open(filename, "rb") as f:
        for line in f:
            if line.decode().startswith('$DefaultValue='):
                default = line.decode().split('=')[1]
                try:
                    return int(default)
                except:
                    return default
        return np.nan


def _get_names(filename):
    """Get variable names from CSV file header.

    Parameters:
        filename (str): Path to CSV file.

    Returns:
        str: Comma separated list of variable names.
    """
    with open(filename, "rb") as f:
        for line in f:
            if line.decode().startswith('$Names='):
                return line.decode().split('=')[1].replace(';', ',').strip()
    raise Exception('No field names were found.')


def _get_skip_header(filename):
    """Get number of lines to skip header of CSV file.

    Parameters:
        filename (str): Path to CSV file.

    Returns:
        int: Number of lines to skip at the beginning of the file.
    """
    i = 0
    with open(filename, 'rb') as f:
        for line in f:
            if line.decode()[0] == '$' or line.decode()[0] == '#':
                i += 1
            else:
                return i
    raise Exception('No valid line found.')


def read(filename, variables=None, delimiter=';', filling_values=np.nan,
         output=None, **kwargs):
    """Read CSV files.

    Parameters:
        filename (str): Path to CSV file.
        variables (List[str]): List of variables to extract.
        delimiter (str): The string used to separate values.
        filling_values (float): Value to use for missing values.
            If `None`, use the `DefaultValue` specified in file header.
        output (dict): Dictionary that is updated with read data.
        **kwargs: Additional keyword arguments passed to `np.genfromtxt`.

    Returns:
        dict: Dictionary containing the data arrays.
    """
    skip_header = _get_skip_header(filename)
    names = _get_names(filename)
    dtype = [(n, VARIABLE_DTYPES.get(n, 'f8')) for n in names.split(',')]
    if filling_values is None:
        filling_values = _get_default_value(filename)

    # Read DATE and TIME even if they are not explicitly listed.
    if variables is not None:
        usecols = list(set(variables + ['DATE', 'TIME']))
    else:
        variables = names.split(',')
        usecols = None

    with open(filename, 'rb') as f:
        data = np.genfromtxt(
            f,
            delimiter=delimiter,
            skip_header=skip_header,
            dtype=dtype,
            names=names,
            filling_values=filling_values,
            usecols=usecols,
            **kwargs,
            )

    # Convert structured array to dictionary.
    if output is None:
        output = {var: data[var] for var in variables}
    else:
        for var in variables:
            output[var] = data[var]

    # Always convert DATE and TIME into matplotlib time.
    dates = [' '.join(d) for d in zip(data['DATE'], data['TIME'])]
    output['MPLTIME'] = _get_mpl_date(dates)

    return output


def read_profile(filename, var_regex=None, var_key='PROFILE', **kwargs):
    """Read scattering coefficients from CSV file.

    Parameters:
        filename (str): Path to CSV file.
        var_regex (str): Python regular expression [0] matching
            the variable name of the profile.
        var_key (str): Dictionary key for extracted profile.
        **kwargs: Additional keyword arguments passed to `read`.

    [0] https://docs.python.org/3.1/library/re.html

    Returns:
        np.ndarray: Profile.
    """
    profile_key = var_key + '_Z'

    output = read(filename, **kwargs)

    p = re.compile(var_regex)

    var_names = [var for var in output.keys() if p.match(var)]
    var_names.sort()

    profile = np.vstack([output[v] for v in var_names])

    z = [float(re.sub('[^0-9]', '', v)) for v in var_names]

    # Extract height information from variable name.
    output[var_key] = np.ma.masked_invalid(profile)
    output[profile_key] = np.array(z)

    return output


def read_scat(filename, var_regex='CLB_B\d{5}', var_key='CLB_MATRIX',
              **kwargs):
    """Read scattering coefficients from CSV file.

    Parameters:
        filename (str): Path to CSV file.
        var_regex (str): Python regular expression [0] matching
            the variable name of the profile.
        var_key (str): Dictionary key for extracted profile.
        **kwargs: Additional keyword arguments passed to `read_profile`.

    [0] https://docs.python.org/3.1/library/re.html

    Returns:
        np.array, np.array: scattering coefficient, height levels
    """
    output = read_profile(
        filename,
        var_key=var_key,
        var_regex=var_regex,
        **kwargs,
        )

    back_scat = output[var_key]
    back_scat = np.ma.masked_less(back_scat, 0)

    output[var_key] = back_scat

    return output


def write(filename, data, variables=None):
    """Write data from dictionary to CSV file.

    Parameters:
        filename (str): Path to CSV file.
        data (dict): Dictionary containing data.
        variables (list[str]): Variables to store.
    """
    if not ('DATE' in data and 'TIME' in data):
        raise Exception('Need DATE and TIME to write proper files.')

    if variables is None:
        variables = list(sorted(data.keys()))
        # Write date and time to first columns.
        for k in ['TIME', 'DATE']:
            if k in variables:
                variables.remove(k)
                variables.insert(0, k)

    # Never write MPLTIME. It is created when reading anyways.
    variables.remove('MPLTIME')

    # Convert dictionart entries to data matrix.
    data_matrix = np.vstack(data[v] for v in variables).T

    # Calculate some time statistics in Wettermast style.
    # It looks messy and it is not clear, if those values are needed.
    # But the results arer equivalent to the original files.
    base_date = '27.03.1995 00:00:00'
    dates = [' '.join(d) for d in zip(data['DATE'], data['TIME'])]
    first_date = dates[0] + ':00'
    mpl_times = _get_mpl_date([base_date, first_date], fmt='%d.%m.%Y %H:%M:%S')
    secs_since = np.diff(mpl_times) * 24 * 3600  # time span in seconds
    time_lag = np.round(np.mean(np.diff(_get_mpl_date(dates) * 24 * 3600)))

    # Combine time properties to file header.
    header = '#={}\n'.format(data_matrix.shape[0])
    header += '$FirstDateTime={}\n'.format(first_date)
    header += '$JSDBaseDateTime={}\n'.format(base_date)
    header += '$FirstJSD={}\n'.format(int(secs_since))
    header += '$TimeLagSec={}\n'.format(int(time_lag))
    header += '$DefaultValue=99999\n'
    header += '$Names=' + ";".join(variables)

    np.savetxt(
        filename,
        data_matrix,
        comments='',
        delimiter=';',
        header=header,
        fmt='%s',
        )
